/** &amp; regex */
const _ar = /&/g;
/** resolve selector */
function r$(amp, selector) {
    if (selector.includes(',')) {
        return selector.split(_sr).map(part => {
            if (_ar.test(part)) {
                return part.replace(_ar, amp);
            }
            else {
                return amp + ' ' + part;
            }
        }).join(', ');
    }
    else if (_ar.test(selector)) {
        return selector.replace(_ar, amp);
    }
    else {
        return amp + ' ' + selector;
    }
}
/** seperator regex */
const _sr = /,(?![^(]*\))/;
/** trim regex */
const _tr = /\s+/g;
/** parse selector path */
function p$p(selectorPaths) {
    let amp = '';
    selectorPaths.forEach(selector => {
        // eslint-disable-next-line
        selector = selector && selector.trim();
        if (
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        !selector) {
            return;
        }
        if (/,/g.test(amp)) {
            amp = amp
                .split(_sr)
                .map(ampPart => r$(ampPart, selector))
                .join(', ');
        }
        else {
            amp = r$(amp, selector);
        }
    });
    return amp.trim().replace(_tr, ' ');
}

/** kebab regex */
const _kr = /[A-Z\u00C0-\u00D6\u00D8-\u00DE]/g;
/** kebab case */
function _kc(pattern) {
    return pattern.replace(_kr, match => '-' + match.toLowerCase());
}
/** upwrap property */
/** TODO: refine it */
function _up(prop, indent = '  ') {
    if (typeof prop === 'object' && prop !== null) {
        return (' {\n' +
            Object.entries(prop).map(v => {
                return indent + `  ${_kc(v[0])}: ${v[1]};`;
            }).join('\n') +
            '\n' + indent + '}');
    }
    return `: ${String(prop)};`;
}
/** unwrap properties */
function _ups(props, instance, params) {
    if (typeof props === 'function') {
        return props({
            context: instance.context,
            props: params
        });
    }
    return props;
}
/** create style */
function _cs(selector, props, instance, params) {
    if (props === null)
        return null;
    // eslint-disable-next-line
    const unwrappedProps = _ups(props, instance, params) || {};
    const propertyNames = Object.keys(unwrappedProps);
    if (propertyNames.length === 0) {
        if (instance.config.preserveEmptyBlock)
            return selector + ' {}';
        return null;
    }
    const statements = [
        selector + ' {'
    ];
    propertyNames.forEach(propertyName => {
        const property = unwrappedProps[propertyName];
        propertyName = _kc(propertyName);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (property !== null && property !== undefined) {
            statements.push(`  ${propertyName}${_up(property)}`);
        }
    });
    statements.push('}');
    return statements.join('\n');
}
/** traverse with callback */
function tc(children, options, callback) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!children)
        return;
    children.forEach(child => {
        if (Array.isArray(child)) {
            tc(child, options, callback);
        }
        else if (typeof child === 'function') {
            const grandChildren = child(options);
            if (Array.isArray(grandChildren)) {
                tc(grandChildren, options, callback);
            }
            else {
                callback(grandChildren);
            }
        }
        else {
            callback(child);
        }
    });
}
/** traverse */
function t(node, selectorPaths, styles, instance, params) {
    const $ = node.$;
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!$ || typeof $ === 'string') {
        // as a string selector
        selectorPaths.push($);
    }
    else if (typeof $ === 'function') {
        // as a lazy selector
        selectorPaths.push($({
            context: instance.context,
            props: params
        }));
    }
    else { // as a option selector
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if ($.before)
            $.before(instance.context);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!$.$ || typeof $.$ === 'string') {
            selectorPaths.push($.$);
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        }
        else if ($.$) {
            selectorPaths.push($.$({
                context: instance.context,
                props: params
            }));
        }
    }
    const selector = p$p(selectorPaths);
    const style = _cs(selector, node.props, instance, params);
    if (style !== null)
        styles.push(style);
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (node.children) {
        tc(node.children, {
            context: instance.context,
            props: params
        }, childNode => {
            t(childNode, selectorPaths, styles, instance, params);
        });
    }
    selectorPaths.pop();
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (typeof $ === 'object' && $ && $.after)
        $.after(instance.context);
}
function render(node, instance, props) {
    const styles = [];
    t(node, [], styles, instance, props);
    return styles.join('\n\n');
}

/* remove el */
function _re(el) {
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (!el)
        return;
    const parentElement = el.parentElement;
    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
    if (parentElement)
        parentElement.removeChild(el);
}
/** query element */
function _qe(id) {
    return document.querySelector(`style[cssr-id="${String(id)}"]`);
}
/** create element */
function _ce(id) {
    const el = document.createElement('style');
    el.setAttribute('cssr-id', String(id));
    return el;
}

/** get count of element */
function _gc(el) {
    const count = el.getAttribute('mount-count');
    if (count === null)
        return 0;
    return Number(count);
}
/** set count of element */
function _sc(el, count) {
    el.setAttribute('mount-count', String(count));
}
/** unmount */
function _u(intance, node, target, count) {
    const els = node.els;
    if (target === undefined) {
        els.forEach(_re);
        node.els = [];
    }
    else if (typeof target === 'string' || typeof target === 'number') {
        const targetElement = _qe(target);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (targetElement && els.includes(targetElement)) {
            const mountCount = _gc(targetElement);
            if (!count || mountCount === 1) {
                _re(targetElement);
                node.els = els.filter(el => el !== targetElement);
            }
            else {
                _sc(targetElement, mountCount - 1);
            }
        }
    }
    else if (els.includes(target)) {
        const mountCount = _gc(target);
        if (!count || mountCount === 1) {
            _re(target);
            node.els = els.filter(el => el !== target);
        }
        else {
            _sc(target, mountCount - 1);
        }
    }
}
/** mount */
function _m(instance, node, target, props, count) {
    let targetElement = null;
    if (target === undefined) {
        targetElement = document.createElement('style');
        document.head.appendChild(targetElement);
    }
    else if (typeof target === 'string' || typeof target === 'number') {
        targetElement = _qe(target);
        if (targetElement === null) {
            targetElement = _ce(target);
            document.head.appendChild(targetElement);
            if (count) {
                _sc(targetElement, 1);
            }
        }
        else {
            if (count) {
                _sc(targetElement, _gc(targetElement) + 1);
            }
            return targetElement;
        }
    }
    else {
        targetElement = target;
        const mountCount = _gc(targetElement);
        if (mountCount > 0) {
            if (count) {
                _sc(targetElement, mountCount + 1);
            }
            return target;
        }
    }
    const style = node.render(props);
    if (targetElement.innerHTML !== style) {
        targetElement.innerHTML = style;
    }
    return targetElement;
}

/** render wrapper */
function _r(props) {
    return render(this, this.instance, props);
}
/** wrapped mount */
function _wm(options) {
    // eslint-disable-next-line
    const target = options && options.target;
    if (target === null)
        return null;
    // eslint-disable-next-line
    const props = options && options.props;
    // eslint-disable-next-line
    const count = !((options && options.count) === false);
    const targetElement = _m(this.instance, this, target, props, count);
    const els = this.els;
    if (!els.includes(targetElement)) {
        els.push(targetElement);
    }
    return targetElement;
}
/** wrapped unmount */
function _wu(options) {
    // eslint-disable-next-line
    const target = options && options.target;
    // eslint-disable-next-line
    const delay = (options && options.delay) || 0;
    // eslint-disable-next-line
    const count = !((options && options.count) === false);
    if (target === null)
        return;
    if (delay === 0)
        _u(this.instance, this, target, count);
    else {
        setTimeout(() => _u(this.instance, this, target, count), delay);
    }
}
/** create CNode */
const _cc = function (instance, $, props, children) {
    return {
        instance,
        $,
        props,
        children,
        els: [],
        render: _r,
        mount: _wm,
        unmount: _wu
    };
};
const c = function (instance, $, props, children) {
    if (Array.isArray($)) {
        return _cc(instance, { $: null }, null, $);
    }
    if (Array.isArray(props)) {
        return _cc(instance, $, null, props);
    }
    else if (Array.isArray(children)) {
        return _cc(instance, $, props, children);
    }
    else {
        return _cc(instance, $, props, null);
    }
};

function CSSRender(config = {
    preserveEmptyBlock: false
}) {
    const cssr = {
        c: ((...args) => c(cssr, ...args)),
        use: (plugin, ...args) => plugin.install(cssr, ...args),
        context: {},
        config
    };
    return cssr;
}

export default CSSRender;
export { CSSRender };
